"""
Generics are defined here to avoid repetition throughout the library
"""

from typing import TypeVar

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")

T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
T7 = TypeVar("T7")
T8 = TypeVar("T8")
T9 = TypeVar("T9")
T10 = TypeVar("T10")
T11 = TypeVar("T11")
T12 = TypeVar("T12")
T13 = TypeVar("T13")
T14 = TypeVar("T14")
T15 = TypeVar("T15")
T16 = TypeVar("T16")

Ret = TypeVar("Ret")

SuccessT = TypeVar("SuccessT")
